// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.


#include <boost/property_tree/json_parser.hpp>
#include <boost/property_tree/ptree.hpp>
#include <chrono>
#include <fstream>
#include <iostream>
#include <string>
#include <sys/file.h>
#include <sys/stat.h>
#include <vector>
#include <xrm.h>
#include <filesystem>

// #define DEBUG

static constexpr uint32_t MAX_REQ_JOB_SLOTS = 9999;
static constexpr uint32_t PLUGIN_COUNT      = 4;
static constexpr uint32_t XRM_PLUGIN_FUNCID = 0;
static constexpr int32_t  XRM_PRECISION_1000000_BIT_MASK(int32_t load) {
  return (load << 8);
}

static constexpr const char* VAR_TMP_AMD                = "/var/tmp/amd/";
static constexpr const char* BASH_FILE_PATH             = "/var/tmp/amd/ma35/";
static constexpr const char* RESERVE_ID_FILENAME_PREFIX = "xrm_jobReservation_";
static constexpr const char* RESERVE_ID_PREFIX          = "export XRM_RESERVE_ID_";
static constexpr const char* DEVICE_ID_PREFIX           = "var_dev_";

namespace pt = boost::property_tree;

struct cuInfo {
  std::vector<uint32_t>    cu_load;
  std::vector<uint32_t>    num_channels;
  std::string              plugin_name;
  std::vector<std::string> kernel_name;
  std::vector<std::string> kernel_alias;
};

struct jobSlotInfo {
  bool                                 is_ull = false;
  uint32_t                             num_dev_per_cmd;
  uint32_t                             req_job_slots = MAX_REQ_JOB_SLOTS;
  uint32_t                             allot_job_slots;
  int32_t                              ref_idx     = -1;
  float                                load_factor = 1.0; // to limit the maximum allowed jobslots per device
  std::string                          json_file_name;
  std::string                          sh_file_name;
  std::ofstream                        sh_fp;
  std::vector<uint32_t>                start_cu_idx;
  std::vector<int32_t>                 reservation_ids;
  std::unique_ptr<xrmCuPoolPropertyV2> xrm_cu_pool_prop;
};

struct jobSlotAppInfo {
  bool                                is_dry_run          = false;
  int32_t                             total_jobs_reserved = 0;
  xrmContext*                         xrm_ctx             = nullptr;
  std::unique_ptr<xrmPluginFuncParam> plugin_param_ptr;
  std::vector<jobSlotInfo>            job_info;
  cuInfo cu_info[PLUGIN_COUNT] = {{{}, {}, "xrmMA35MLPlugin", {"ml"}, {"ML_AMA"}}, {{}, {}, "xrmMA35ScalePlugin", {"scaler"}, {"SCALER_AMA"}},
      {{}, {}, "xrmMA35DecodePlugin", {"decoder"}, {"DECODER_AMA"}},
      {{}, {}, "xrmMA35EncodePlugin", {"encoder", "encoder", "lookahead"}, {"ENCODER_TYPE1_AMA", "ENCODER_TYPE2_AMA", "LOOKAHEAD_AMA"}}};

  ~jobSlotAppInfo() {
    // Remove bash files created to store reserve Ids if not a dry-run
    if (!is_dry_run) {
      for (size_t file = 0; file < job_info.size(); ++file) {
        if (job_info[file].sh_fp && job_info[file].ref_idx < 0)
          if (remove(job_info[file].sh_file_name.c_str()) == 0)
            fprintf(stdout, "\n%s : Deleted successfully\n", job_info[file].sh_file_name.c_str());
      }
    }
    // relinquish reserve Ids and free free correspsonding CU reources
    for (size_t files = 0; files < job_info.size(); ++files) {
      for (size_t i = 0; i < job_info[files].reservation_ids.size(); ++i) {
        xrmCuPoolRelinquish(xrm_ctx, job_info[files].reservation_ids[i]);
      }
    }
    // Destroy XRM context
    xrmDestroyContext(xrm_ctx);
  };
};

#ifdef DEBUG
// Copied from https://gitenterprise.xilinx.com/FaaSApps/XRM/blob/ma35_support/test/example_1/src/example_test_xrm_api.c
// Dont modify till the XRM development completes for MA35
static void printCuResourceInfoV2(xrmCuResourceV2 cuRes) {
  // printf("   xclbinFileName is:  %s\n", cuRes.xclbinFileName);
  // printf("   kernelPluginFileName is:  %s\n", cuRes.kernelPluginFileName);
  printf("   kernelName is:  %s\n", cuRes.kernelName);
  printf("   instanceName is:  %s\n", cuRes.instanceName);
  printf("   cuName is:  %s\n", cuRes.cuName);
  printf("   kernelAlias is:  %s\n", cuRes.kernelAlias);
  printf("   deviceNumBDF is:  0x%x\n", cuRes.deviceNumBDF);
  printf("   cuId is:  %d\n", cuRes.cuId);
  printf("   channelId is:  %d\n", cuRes.channelId);
  printf("   cuType is:  %d\n", cuRes.cuType);
  // printf("   baseAddr is:  0x%lx\n", cuRes.baseAddr);
  // printf("   membankId is:  %d\n", cuRes.membankId);
  // printf("   membankType is:  %d\n", cuRes.membankType);
  // printf("   membankSize is:  0x%lx\n", cuRes.membankSize);
  // printf("   membankBaseAddr is:  0x%lx\n", cuRes.membankBaseAddr);
  printf("   allocServiceId is:  %lu\n", cuRes.allocServiceId);
  printf("   poolId is:  %lu\n", cuRes.poolId);
  printf("   channelLoad is:  %d\n", cuRes.channelLoad >> 8);
  if (cuRes.logicGroupInfor.indicator == XRM_LOGIC_GROUP_INFO_ID_TYPE_VALID) {
    printf("   logic gropu infor: indicator:  %d\n", cuRes.logicGroupInfor.indicator);
    printf("   logic gropu infor: groupId:  %d\n", cuRes.logicGroupInfor.groupId);
    // printf("   logic gropu infor: groupType:  %s\n", cuRes.logicGroupInfor.groupType);
    printf("   sliceId:  %d\n", (cuRes.logicGroupInfor.groupId > 1) ? 1 : 0);
  }
  printf("   deviceId is:  %d\n", cuRes.deviceId);
}
#endif /* DEBUG */

// print -h and --help options when called
static void print_help(const char** argv) {
  fprintf(stdout, "Options available:\n");
  fprintf(stdout,
      " -h, --help            Show this help message and exit\n -d, --dry_run         Set dry run enable, Only display no of slots possible for given job_description "
      "files.\n\nUsage:-\nFor Reservation:\n");
  fprintf(stdout, "    		%s file_1.json [file_2.json] ... [file_n.json]\n", argv[0]);
  fprintf(stdout, "For Dry run (to know how many job slots are possible):\n");
  fprintf(stdout, "    		%s file_1.json [file_2.json] ... [file_n.json] --dry_run\n", argv[0]);
  return;
}

// check access for bash file creation
static int check_folder_access() {
  struct stat dstat;
  if (stat(BASH_FILE_PATH, &dstat) == -1) {
    mode_t  target_mode = S_IRWXU | S_IRWXG | S_IRWXO;
    int32_t ret         = mkdir(VAR_TMP_AMD, target_mode);
    if (ret == EEXIST) {
      fprintf(stdout, "\nFolder exists skiping mkdir\n");
    } else if (ret != 0) {
      fprintf(stderr, "\nError: Couldn't create %s folder to write logs. Err = %s\n", VAR_TMP_AMD, strerror(errno));
      return 1;
    } else {
      // Fixed folder access issue for multi-user
      chmod(VAR_TMP_AMD, target_mode);
    }
    ret = mkdir(BASH_FILE_PATH, target_mode);
    if (ret == EEXIST) {
      fprintf(stdout, "\nFolder exists skiping mkdir\n");
    } else if (ret != 0) {
      fprintf(stderr, "\nError: Couldn't create %s folder to write logs. Err = %s\n", BASH_FILE_PATH, strerror(errno));
      return 1;
    } else {
      // Fixed folder access issue for multi-user
      chmod(BASH_FILE_PATH, target_mode);
    }
  }
  return 0;
}

// Check Valid or invalid arguments in argc
static int parse_arguments(uint32_t argc, const char** argv, jobSlotAppInfo* app_info_ptr) {
  if (argc < 2) {
    print_help(argv);
    return 1;
  }

  for (uint32_t i = 1, j = 0; i < argc; ++i) {
    if (argv[i][0] == '-') {
      if ((argv[i][1] == 'd' && argv[i][2] == '\0') || strcmp("--dry_run", argv[i]) == 0)
        app_info_ptr->is_dry_run = true;
      else if ((argv[i][1] == 'h' && argv[i][2] == '\0') || strcmp("--help", argv[i]) == 0) {
        print_help(argv);
        return 1;
      } else {
        fprintf(stderr, "\nError: Unknown Option %s\nGet info on usage:\n     %s --help \n   or   \n     %s -h \n", argv[i], argv[0], argv[0]);
        return 1;
      }
    } else {
      std::string file_path(argv[i]);
      if (file_path.find(".json") != std::string::npos) {
        std::ifstream file(file_path);
        if (!file.is_open()) {
          fprintf(stderr, "\nError: %s file not found\n", file_path.c_str());
          return 1;
        }
        file.close();
        app_info_ptr->job_info.push_back(jobSlotInfo());
        app_info_ptr->job_info[j++].json_file_name = argv[i];
      } else {
        fprintf(stderr, "\nError: Unknown file-Fomat %s\nGet info on usage:\n     %s --help \n   or   \n     %s -h \n", argv[i], argv[0], argv[0]);
        return 1;
      }
    }
  }

  if (app_info_ptr->job_info.size() < 1) {
    fprintf(stderr, "\nError: No Job_description files provided, exiting......\n");
    return 1;
  }

  // Check access to "/var/tmp/amd" folder to xrmReservation Ids in SH files
  if (!app_info_ptr->is_dry_run && check_folder_access())
    return 1;

#ifdef DEBUG
  fprintf(stdout,
      "\n###############################################################"
      "\n\n");
#endif

  return 0;
}

// Get fileName without extension and path
static __inline std::string getFileName(std::string file_path) {
  std::filesystem::path fs_path(file_path);
  return fs_path.stem();
}

// This function needs to be called once for each CU request from XRM, which populates the XRM CU properties for the CU.
static void fill_xrm_cu_props(
    xrmCuPoolPropertyV2* xrm_cu_pool_prop, uint32_t cu_load, uint32_t cu_index, const std::string& kernel_name, const std::string& kernel_alias, uint32_t dev_idx) {
  uint64_t deviceInfoContraintType = XRM_DEVICE_INFO_CONSTRAINT_TYPE_VIRTUAL_DEVICE_INDEX;
  strcpy(xrm_cu_pool_prop->cuListProp.cuProps[cu_index].kernelName, kernel_name.c_str());
  strcpy(xrm_cu_pool_prop->cuListProp.cuProps[cu_index].kernelAlias, kernel_alias.c_str());
  xrm_cu_pool_prop->cuListProp.cuProps[cu_index].devExcl    = false;
  xrm_cu_pool_prop->cuListProp.cuProps[cu_index].deviceInfo = (dev_idx << XRM_DEVICE_INFO_DEVICE_INDEX_SHIFT) |
      (deviceInfoContraintType << XRM_DEVICE_INFO_CONSTRAINT_TYPE_SHIFT);
  xrm_cu_pool_prop->cuListProp.cuProps[cu_index].requestLoad = XRM_PRECISION_1000000_BIT_MASK(cu_load);
  return;
}

// This function poupulates the Logical Grouping Information needed to support Slice Id, Slice Pinning & ULL cases.
// Needed to request VC8000E, XAV1 Encoder and corresponding Lookahead CU allocations from XRM
static void update_lg_info(xrmCuPoolPropertyV2* xrm_cu_pool_prop, uint32_t cu_load, bool is_ull, uint32_t& cu_index, const std::string& kernel_alias,
    uint32_t lookahead_load, int32_t pseudo_logic_group_idx) {
  // As per the XRM 2nd drop design, constraint info should be set for each CU to meet Slice aware, Slice Pinning & ULL requirements
  // For ULL, XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_REAL_LOGIC_GROUP_ID should be set with actual groupId
  // For Slice pinning of Encoder and Lookahead, XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_PSEUDO_LOGIC_GROUP_ID should be set with psuedo groupId (only if Lookahead IP is enabled)
  // Slice Aware, XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_LOGIC_GROUP_TYPE should be used with appropriate groupType (lookahead is disabled)
  uint32_t constraint_info                                                     = is_ull ?
                                                          XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_REAL_LOGIC_GROUP_ID :
                                                          (lookahead_load ? XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_PSEUDO_LOGIC_GROUP_ID : XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_LOGIC_GROUP_TYPE);
  xrm_cu_pool_prop->cuListProp.cuProps[cu_index].logicGroupInfo.constraintInfo = constraint_info;
  int32_t real_logic_group_idx                                                 = 0;
  xrm_cu_pool_prop->cuListProp.cuProps[cu_index].logicGroupInfo.groupId        = is_ull ? real_logic_group_idx : pseudo_logic_group_idx;
  // Group Type is valid only if constraint_info = XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_LOGIC_GROUP_TYPE
  if (xrm_cu_pool_prop->cuListProp.cuProps[cu_index].logicGroupInfo.constraintInfo == XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_LOGIC_GROUP_TYPE)
    strcpy(xrm_cu_pool_prop->cuListProp.cuProps[cu_index].logicGroupInfo.groupType, (kernel_alias == "ENCODER_TYPE1_AMA") ? "ENCODER_TYPE1" : "ENCODER_TYPE2");

  // Request for additional CUs of Encoder & lookahead from Slice 1 if it is a ULL encode channel
  if (is_ull) {
    // make 2 half loads and query from both slices for ULL
    cu_load = (cu_load + 1) >> 1;
    // Overwrite with updated cu_load
    xrm_cu_pool_prop->cuListProp.cuProps[cu_index].requestLoad = XRM_PRECISION_1000000_BIT_MASK(cu_load);
    if (kernel_alias == "ENCODER_TYPE1_AMA") {
      // Default groupID is zero. increment by 1, if it is XAV1 encoder CU.
      xrm_cu_pool_prop->cuListProp.cuProps[cu_index].logicGroupInfo.groupId++;
    } else if (kernel_alias == "LOOKAHEAD_AMA") {
      // Lookahead CU group ID should be same as corresponding Encoder CU group ID. So populating from previously populated Encoder CU properties.
      xrm_cu_pool_prop->cuListProp.cuProps[cu_index].logicGroupInfo.groupId = xrm_cu_pool_prop->cuListProp.cuProps[cu_index - 2].logicGroupInfo.groupId;
    }
    memcpy(&xrm_cu_pool_prop->cuListProp.cuProps[cu_index + 1], &xrm_cu_pool_prop->cuListProp.cuProps[cu_index], sizeof(xrmCuPropertyV2));
    xrm_cu_pool_prop->cuListProp.cuProps[cu_index + 1].logicGroupInfo.groupId += 2;
    // fprintf (stdout,  "\nGroup id 0: %d\n", xrm_cu_pool_prop->cuListProp.cuProps[cu_index].logicGroupInfo.groupId);
    // fprintf (stdout,  "\nGroup id 1: %d\n", xrm_cu_pool_prop->cuListProp.cuProps[cu_index+1].logicGroupInfo.groupId);
    cu_index += 1;
  }

  return;
}

// Populate XRM cu pool property structs for non zero values of CU loads
static void fill_xrm_props(jobSlotAppInfo* app_info_ptr, int32_t job_index) {
  uint32_t index                                                = 0;
  int32_t  enc_load_idx                                         = 0;
  app_info_ptr->job_info[job_index].xrm_cu_pool_prop->cuListNum = 1;
  int32_t pseudo_logic_group_idx                                = -1;

  // Device loop
  for (uint32_t dev_idx = 0; dev_idx < app_info_ptr->job_info[job_index].num_dev_per_cmd; ++dev_idx) {
    app_info_ptr->job_info[job_index].start_cu_idx.push_back(index);
    // Plugin loop
    for (size_t plugin_idx = 0; plugin_idx < sizeof(app_info_ptr->cu_info) / sizeof(cuInfo); ++plugin_idx) {
      cuInfo* job_cu_info   = app_info_ptr->cu_info;
      bool    is_encoder_cu = (job_cu_info[plugin_idx].kernel_name[0] == "encoder");
      // Channel loop > 1 valid for Encoder CUs only
      for (uint32_t channel_count = 0; channel_count < job_cu_info[plugin_idx].num_channels[dev_idx]; ++channel_count) {
        // load_idx is same as dev_idx, except for encoder CU
        int32_t load_idx = (is_encoder_cu) ? enc_load_idx : dev_idx;
        bool    is_xav1  = false;
        // job_cu_info[plugin_idx].kernel_name.size() is 3 for encoder CU and 1 for rest of CUs
        for (size_t i = 0; i < job_cu_info[plugin_idx].kernel_name.size(); ++i) {
          // enc_load is a 3 value pair (is_xav1 flag, Encoder Load, LA load)
          if (is_encoder_cu && 0 == i) {
            is_xav1 = job_cu_info[plugin_idx].cu_load[load_idx + i];
            continue;
          }
          // populate cu info details for a non zero load
          if (job_cu_info[plugin_idx].cu_load[load_idx + i] > 0) {
            fprintf(stdout, "\n%s [%u]: %u", job_cu_info[plugin_idx].kernel_name[i].c_str(), channel_count, job_cu_info[plugin_idx].cu_load[load_idx + i]);
            fill_xrm_cu_props(app_info_ptr->job_info[job_index].xrm_cu_pool_prop.get(), job_cu_info[plugin_idx].cu_load[load_idx + i], index,
                job_cu_info[plugin_idx].kernel_name[i - is_xav1], job_cu_info[plugin_idx].kernel_alias[i - is_xav1], dev_idx);
            // XRM 2nd drop integration - Logical Group Indexing support
            // Reset LG contraint info
            app_info_ptr->job_info[job_index].xrm_cu_pool_prop->cuListProp.cuProps[index].logicGroupInfo.constraintInfo = XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_NULL;
            // update LG info for Encoder CUs only
            if (is_encoder_cu) {
              update_lg_info(app_info_ptr->job_info[job_index].xrm_cu_pool_prop.get(), job_cu_info[plugin_idx].cu_load[load_idx + i],
                  app_info_ptr->job_info[job_index].is_ull, index, job_cu_info[plugin_idx].kernel_alias[i - is_xav1], job_cu_info[plugin_idx].cu_load[load_idx + 2],
                  pseudo_logic_group_idx);
            }
            is_xav1 = false;
            ++index;
          }
        }
        --pseudo_logic_group_idx;

        if (is_encoder_cu)
          enc_load_idx += job_cu_info[plugin_idx].kernel_name.size();
      }
    }
  }
  app_info_ptr->job_info[job_index].xrm_cu_pool_prop->cuListProp.cuNum = index;
  return;
}

// Performs load calculation for a given job type
static int calc_job_load(jobSlotAppInfo* app_info_ptr, int32_t job_index, pt::ptree job) {
  xrmPluginFuncParam* param = app_info_ptr->plugin_param_ptr.get();
  memset(param, 0, sizeof(xrmPluginFuncParam));
  std::stringstream job_str;
  boost::property_tree::write_json(job_str, job);

  // Copy json string(job info) into param->input to send it to XRM plugins
  strncpy(param->input, job_str.str().c_str(), job_str.str().length());

  // Check for each plugin at a time
  for (uint32_t plugin_idx = 0; plugin_idx < PLUGIN_COUNT; ++plugin_idx) {
    cuInfo* job_cu_info = app_info_ptr->cu_info;

    // Reset cu_load and num_channels for next job
    job_cu_info[plugin_idx].cu_load.clear();
    job_cu_info[plugin_idx].num_channels.clear();

    // Invoke the respective XRM plugin for load calculation
    if (xrmExecPluginFunc(app_info_ptr->xrm_ctx, const_cast<char*>(job_cu_info[plugin_idx].plugin_name.c_str()), XRM_PLUGIN_FUNCID, param) != XRM_SUCCESS) {
      fprintf(stderr, "\nError: %s plugin function= %u fail to run the function\n", job_cu_info[plugin_idx].kernel_name[0].c_str(), XRM_PLUGIN_FUNCID);
      return 1;
    } else {
      char* endptr;
      char* token;
#ifdef DEBUG
      fprintf(stdout, "\nPlugin : %s\n", job_cu_info[plugin_idx].kernel_name[0].c_str());
      fprintf(stdout, "\nOutput param is : %s\n", param->output);
#endif
      errno = 0;
      token = static_cast<char*>(strtok(param->output, " "));
      if (token == NULL)
        break;

      app_info_ptr->job_info[job_index].num_dev_per_cmd = static_cast<uint32_t>(strtol(token, &endptr, 0));
      if (errno != 0) {
        fprintf(stderr, "\nError: String to long conversion error / Load token not found\n");
        return 1;
      }

      // loop count indicates device count
      for (uint32_t nd = 0; nd < app_info_ptr->job_info[job_index].num_dev_per_cmd; ++nd) {
        uint32_t num_channels = static_cast<int32_t>(strtol(static_cast<char*>(strtok(NULL, " ")), &endptr, 0));
        if (errno != 0) {
          fprintf(stderr, "\nError: String to long conversion error / no of channels token not found \n");
          return 1;
        }

        job_cu_info[plugin_idx].num_channels.push_back(num_channels);

        // Loop count indicates no of channels for each plugin type
        for (uint32_t i = 0; i < num_channels; ++i) {
          // Channel loop > 1 valid for Encoder CUs only
          for (size_t j = 0; j < job_cu_info[plugin_idx].kernel_name.size(); ++j) {
            // get the CU loads
            uint32_t cu_load = static_cast<int32_t>(strtol(static_cast<char*>(strtok(NULL, " ")), &endptr, 0));
            cu_load          = cu_load * app_info_ptr->job_info[job_index].load_factor;
            // check for strtol errors
            if (errno != 0) {
              fprintf(stderr, "\nError: String to long conversion error / Load token not found\n");
              return 1;
            }
            if (cu_load > XRM_MAX_CU_LOAD_GRANULARITY_1000000 && !(app_info_ptr->job_info[i].is_ull && job_cu_info[plugin_idx].kernel_name[0] == "encoder")) {
              fprintf(stderr, "\nError: Requested %s load = %u exceeds maximum capacity.\n", job_cu_info[plugin_idx].kernel_name[0].c_str(), cu_load);
              return 1;
            }
            job_cu_info[plugin_idx].cu_load.push_back(cu_load);
          }
        }
      }
    }
  }
  // Fill xrm cu pool property structure as per the CU loads
  fill_xrm_props(app_info_ptr, job_index);
  return 0;
}

// Performs load calculation for all job types
static int calc_all_job_load(jobSlotAppInfo* app_info_ptr) {
  // loop over job types - heterogenous support
  for (size_t i = 0; i < app_info_ptr->job_info.size(); ++i) {
    // Parse num_job_slots value if present in json file
    pt::ptree job;
    try {
      pt::read_json(app_info_ptr->job_info[i].json_file_name.c_str(), job);
      // Overwrite num_job_slots variable, if num_job_slots specified explicitly in json
      auto pr_ptree = job.get_child("request.parameters");
      if (pr_ptree.not_found() != pr_ptree.find("num_job_slots"))
        app_info_ptr->job_info[i].req_job_slots = pr_ptree.get<int32_t>("num_job_slots");
      // Overwrite if specified load_factor explicitly - to limit maximum allowed jobslots per device
      if (pr_ptree.not_found() != pr_ptree.find("load_factor")) {
        float load_factor = pr_ptree.get<float>("load_factor");
        // not allowing load_factor < 1.0 which will lead to over subscription
        if (load_factor > 1.0f) {
          app_info_ptr->job_info[i].load_factor = load_factor;
        } else if (load_factor < 1.0f) {
          fprintf(stdout, "Found load_factor(%0.2f) < 1.0f, defaults to 1.0f\n", load_factor);
        }
      }
      // For GA1.0, dual-pass, single slice ULL only supported and the resource requirements this mode is same Normal latency.
      // So no special handling is needed for resource management
      // if (pr_ptree.not_found() != pr_ptree.find("latency_mode") && strcmp("ull", pr_ptree.get<std::string>("latency_mode").c_str()) == 0)
      //   app_info_ptr->job_info[i].is_ull = true;
    } catch (std::exception& e) {
      fprintf(stderr, "\nError: Exception %s While reading json File\n", e.what());
      return 1;
    }
    fprintf(stdout, "\nJob Description File: %s\n", app_info_ptr->job_info[i].json_file_name.c_str());
    if (app_info_ptr->job_info[i].load_factor > 1.0f) {
      fprintf(stdout, "Load Factor: %0.2f applied on calculated CU loads to limit maximum possible jobslots per device\n", app_info_ptr->job_info[i].load_factor);
    }
    fprintf(stdout, "Requested Loads:");
    // Calculate load for each job type
    if (calc_job_load(app_info_ptr, i, job)) {
      fprintf(stderr, "\nError: Job load calculation error in %s", app_info_ptr->job_info[i].json_file_name.c_str());
      return 1;
    }
  }
#ifdef DEBUG
  fprintf(stdout,
      "\n###############################################################"
      "\n\n");
#endif
  return 0;
}

// Check if arguement is repeated (to get ratio based/priority based allocations) and get the initil reference idx
static int check_duplicate_arg(jobSlotAppInfo* app_info_ptr, uint32_t index_to_check) {
  for (uint32_t i = 0; i < index_to_check; ++i) {
    if (app_info_ptr->job_info[i].json_file_name == app_info_ptr->job_info[index_to_check].json_file_name) {
      app_info_ptr->job_info[index_to_check].ref_idx = i;
      return 0;
    }
  }

  return 1;
}

static int initialize_jobslot_app(jobSlotAppInfo* app_info_ptr) {
  // Create XRM context
  app_info_ptr->xrm_ctx = static_cast<xrmContext*>(xrmCreateContext(XRM_API_VERSION_1));
  if (nullptr == app_info_ptr->xrm_ctx) {
    fprintf(stderr, "\nError: Xrm context creation failed\n");
    return 1;
  }

  // Allocate memory for plugin input & output params
  try {
    app_info_ptr->plugin_param_ptr = std::make_unique<xrmPluginFuncParam>();
  } catch (...) {
    fprintf(stderr, "\nError: memory allocation failed for plugin_param_ptr");
    return 1;
  }

  for (size_t i = 0; i < app_info_ptr->job_info.size(); ++i) {
    try {
      app_info_ptr->job_info[i].xrm_cu_pool_prop = std::make_unique<xrmCuPoolPropertyV2>();
    } catch (...) {
      fprintf(stderr, "\nError: memory allocation failed for xrm_cu_pool_prop\n");
      return 1;
    }
    memset(app_info_ptr->job_info[i].xrm_cu_pool_prop.get(), 0, sizeof(xrmCuPoolPropertyV2));

    // Create SH filename with time stamp and keep it ready for write
    if (!check_duplicate_arg(app_info_ptr, i) && i != 0) {
      continue;
    } else {
      uint64_t ns = (std::chrono::duration_cast<std::chrono::nanoseconds>(std::chrono::system_clock::now().time_since_epoch()).count()) % (1000 * 1000 * 100);
      app_info_ptr->job_info[i].sh_file_name = app_info_ptr->job_info[i].sh_file_name + BASH_FILE_PATH + RESERVE_ID_FILENAME_PREFIX + std::to_string(ns) + "_" +
          getFileName(app_info_ptr->job_info[i].json_file_name) + ".sh";
      if (!app_info_ptr->is_dry_run) {
        app_info_ptr->job_info[i].sh_fp.open(app_info_ptr->job_info[i].sh_file_name, std::ios::out);
        if (!app_info_ptr->job_info[i].sh_fp) {
          fprintf(stderr, "\nError: Couldn't create SH file: %s \n", app_info_ptr->job_info[i].sh_file_name.c_str());
          return 1;
        }
      }
    }
  }
  return 0;
}

// Update reservation Id to the sh file created
static int update_xrm_reservation_ids(jobSlotAppInfo* app_info_ptr) {
  uint32_t                             total_pending_job_types                          = app_info_ptr->job_info.size();
  int32_t                              is_reservation_complete[total_pending_job_types] = {0};
  std::string                          xrm_ch_id, xrm_dev_id;
  xrmCuResInforV2*                     cu_res;
  xrmCuListResInforV2*                 cu_list_res;
  std::unique_ptr<xrmCuPoolResInforV2> xrm_cu_pool_res;
  try {
    xrm_cu_pool_res = std::make_unique<xrmCuPoolResInforV2>();
  } catch (...) {
    fprintf(stderr, "\nError: memory allocation failed for app_info_ptr\n");
    return 1;
  }

  while (true) {
    for (size_t i = 0; i < app_info_ptr->job_info.size(); ++i) {
      xrm_ch_id.clear();
      if (1 == is_reservation_complete[i])
        continue;
      int32_t reservation_id = xrmCuPoolReserveV2(app_info_ptr->xrm_ctx, app_info_ptr->job_info[i].xrm_cu_pool_prop.get(), xrm_cu_pool_res.get());
      if (reservation_id > 0) {
#ifdef DEBUG
        {
          fprintf(stdout, "JOB SLOT %d\n", app_info_ptr->total_jobs_reserved);
          for (int m = 0; m < app_info_ptr->job_info[i].xrm_cu_pool_prop->cuListProp.cuNum; m++) {
            xrmCuListPropertyV2 cu_list_prop;
            int                 ret = -1;
            memset(&cu_list_prop, 0, sizeof(xrmCuListPropertyV2));
            xrmCuListResourceV2 cu_list_res2;
            memset(&cu_list_res2, 0, sizeof(xrmCuListResourceV2));

            cu_list_prop.cuNum = 1;
            strcpy(cu_list_prop.cuProps[0].kernelName, app_info_ptr->job_info[i].xrm_cu_pool_prop->cuListProp.cuProps[m].kernelName);
            strcpy(cu_list_prop.cuProps[0].kernelAlias, app_info_ptr->job_info[i].xrm_cu_pool_prop->cuListProp.cuProps[m].kernelAlias);
            cu_list_prop.cuProps[0].devExcl     = false;
            cu_list_prop.cuProps[0].requestLoad = (app_info_ptr->job_info[i].xrm_cu_pool_prop->cuListProp.cuProps[m].requestLoad);
            cu_list_prop.cuProps[0].poolId      = reservation_id;

            ret = xrmCuListAllocV2(app_info_ptr->xrm_ctx, &cu_list_prop, &cu_list_res2);

            if (ret != 0) {
              fprintf(stderr, "xrm_allocation: fail to allocate cu list from reserve id : %d\n", reservation_id);
              return ret;
            }
            fprintf(stdout, "\n   CuNum : %d\n", m);
            printCuResourceInfoV2(cu_list_res2.cuResources[0]);
          }
          fprintf(stdout, "\n#####################################################\n\n");
        }
#endif /* DEBUG */
        app_info_ptr->job_info[i].reservation_ids.push_back(reservation_id);
        if (!app_info_ptr->is_dry_run) {
          xrm_ch_id = xrm_ch_id + RESERVE_ID_PREFIX + std::to_string(app_info_ptr->total_jobs_reserved) + "=" +
              std::to_string(app_info_ptr->job_info[i].reservation_ids[app_info_ptr->job_info[i].reservation_ids.size() - 1]) + "\n";
          if (app_info_ptr->job_info[i].ref_idx >= 0) {
            app_info_ptr->job_info[app_info_ptr->job_info[i].ref_idx].sh_fp << xrm_ch_id;
          } else
            app_info_ptr->job_info[i].sh_fp << xrm_ch_id;
          if (app_info_ptr->job_info[i].num_dev_per_cmd > 1) {
            xrm_dev_id.clear();
            for (uint32_t nd = 0; nd < app_info_ptr->job_info[i].num_dev_per_cmd; ++nd) {
              cu_list_res = &(xrm_cu_pool_res->cuListResInfor[0]);
              cu_res      = &(cu_list_res->cuResInfor[app_info_ptr->job_info[i].start_cu_idx[nd]]);
              xrm_dev_id  = xrm_dev_id + DEVICE_ID_PREFIX + std::to_string(app_info_ptr->total_jobs_reserved) + "_" + std::to_string(nd) + "=" +
                  std::to_string(cu_res->deviceId) + "\n";
            }
            xrm_dev_id = xrm_dev_id + "\n";
            if (app_info_ptr->job_info[i].ref_idx >= 0)
              app_info_ptr->job_info[app_info_ptr->job_info[i].ref_idx].sh_fp << xrm_ch_id;
            else
              app_info_ptr->job_info[i].sh_fp << xrm_dev_id;
          }
        }
        ++app_info_ptr->job_info[i].allot_job_slots;
        ++app_info_ptr->total_jobs_reserved;
      }
      if (app_info_ptr->job_info[i].allot_job_slots == app_info_ptr->job_info[i].req_job_slots || reservation_id < 1) {
        if (!is_reservation_complete[i]) {
          is_reservation_complete[i] = 1;
          --total_pending_job_types;
        }
      }
    }
    if (total_pending_job_types == 0)
      break;
  }
  for (size_t j = 0; j < app_info_ptr->job_info.size(); ++j) {
    if (!app_info_ptr->is_dry_run && app_info_ptr->job_info[j].ref_idx < 0)
      app_info_ptr->job_info[j].sh_fp.close();
  }

  return 0;
}

// Validates reservation status and exit for fail cases
static int validate_reservation_status(jobSlotAppInfo* app_info_ptr) {
  fprintf(stdout, "\n===============================================================\n");
  fprintf(stdout, "Total Job Slots possible : %d\n", app_info_ptr->total_jobs_reserved);
  fprintf(stdout, "Dry run: %s\n", (app_info_ptr->is_dry_run ? "Enabled" : "Disabled"));

  // Return with non-zero, if "Job Slots requested" != "Job Slots alloted", valid for explicit job slot request
  for (size_t i = 0; i < app_info_ptr->job_info.size(); ++i) {
    if (app_info_ptr->job_info[i].req_job_slots != MAX_REQ_JOB_SLOTS && (app_info_ptr->job_info[i].req_job_slots > app_info_ptr->job_info[i].allot_job_slots)) {
      fprintf(stderr, " Error: Job Slots requested (%u) >= : Job Slots alloted (%u)", app_info_ptr->job_info[i].req_job_slots, app_info_ptr->job_info[i].allot_job_slots);
      return 1;
    }
  }
  for (size_t j = 0; j < app_info_ptr->job_info.size(); ++j) {
    if (app_info_ptr->job_info[j].ref_idx >= 0) {
      app_info_ptr->job_info[app_info_ptr->job_info[j].ref_idx].allot_job_slots += app_info_ptr->job_info[j].allot_job_slots;
    }
  }

  // Print all info for a each Job_description file provided
  for (size_t k = 0; k < app_info_ptr->job_info.size(); ++k) {
    if (app_info_ptr->job_info[k].ref_idx < 0) {
      fprintf(stdout, "\nJob Type: %s\n", app_info_ptr->job_info[k].json_file_name.c_str());
      if (app_info_ptr->job_info[k].req_job_slots != MAX_REQ_JOB_SLOTS && !(app_info_ptr->job_info[k].req_job_slots < app_info_ptr->job_info[k].allot_job_slots))
        fprintf(stdout, "Job Slots Requested: %u\n", app_info_ptr->job_info[k].req_job_slots);
      fprintf(stdout, "Job Slots Alloted: %u\n", app_info_ptr->job_info[k].allot_job_slots);
      if (!app_info_ptr->is_dry_run && app_info_ptr->job_info[k].allot_job_slots) {
        fprintf(stdout, "XRM_RESERVE_ID file - \"%s\"\n", app_info_ptr->job_info[k].sh_file_name.c_str());
      }
    }
  }
  fprintf(stdout, "================================================================\n");
  // Return with non-zero, if 0 slots are possible
  if (!app_info_ptr->total_jobs_reserved) {
    fprintf(stderr, "Error: Total job slots alloted: %d\n", app_info_ptr->total_jobs_reserved);
    return 1;
  }
  return 0;
}

int main(int argc, const char** argv) {
  std::unique_ptr<jobSlotAppInfo> app_info_ptr;
  try {
    app_info_ptr = std::make_unique<jobSlotAppInfo>();
  } catch (...) {
    fprintf(stderr, "\nError: memory allocation failed for JobSlot info ptr\n");
    return EXIT_FAILURE;
  }

  // Parse & Validate command-line arguments
  if (parse_arguments(argc, argv, app_info_ptr.get())) {
    return EXIT_FAILURE;
  }

  // Initilaize data structures needed and create XRM context, reservation Id
  // bash files for each job type
  if (initialize_jobslot_app(app_info_ptr.get())) {
    return EXIT_FAILURE;
  }

  // Calculate load for each job type
  if (calc_all_job_load(app_info_ptr.get())) {
    return EXIT_FAILURE;
  }

  // update resevation id on sh files
  if (update_xrm_reservation_ids(app_info_ptr.get())) {
    return EXIT_FAILURE;
  }

  // Get Reseravation information and error checking for slots availablity.
  if (validate_reservation_status(app_info_ptr.get())) {
    return EXIT_FAILURE;
  }

  if (!app_info_ptr->is_dry_run) {
    fprintf(stdout,
        "\n---------------------------------------------------------------\n"
        "\nThe Job-slot reservations are alive as long as this Application is"
        " alive!\n(press Enter to close this app)\n\n"
        "---------------------------------------------------------------\n");
    while (true) {
      char endKey = getchar();
      if ((endKey == 13) || (endKey == '\n')) {
        break;
      }
    }
  }

  return 0;
}
